#!/usr/bin/env python
# Copyright (C) 2007 Edgar Merino <donvodka at gmail dot com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# TODO implement option "remove duplicates" to the edit menu in exaile

from UserDict import UserDict
import sys, os, gtk, time, subprocess, tempfile
# if used outside exaile, import threading
# import threading
# exaile imports
from xl import common, xlmisc
from gettext import gettext as _
import xl.plugins as plugins

try:
    import shout
    SHOUT_AVAIL = True
except ImportError:
    SHOUT_AVAIL = False

# Code taken from exaile src
# if you want to reuse the code, you'll need this
#def threaded(f):
#    """
#        A decorator that will make any function run in a new thread
#    """
#    def wrapper(*args, **kwargs):
#        t = threading.Thread(target=f, args=args, kwargs=kwargs)
#        t.setDaemon(True)
#        t.start()
#
#    wrapper.__name__ = f.__name__
#    wrapper.__dict__ = f.__dict__
#    wrapper.__doc__ = f.__doc__
#
#    return wrapper

# Reusable abstract class
# This is needed to make the class Icast not depend on exaile
class Client:
    """
        Main Client interface
        Other clients will subclass it
    """
    def __init__(self, app):
        self.app = app # player that holds a reference to the current song being played
        self.current_track = None # Our players current track

    def is_playing(self):
		return False

    def get_current_track(self):
   		return None

    def get_next_track(self):
        return None

    def get_track_metadata(self):
        """
            Metadata to send to the server
        """
        return self.get_current_track()

    def close(self):
        """
            Do any operations needed when the client is closed
        """
        self.current_track = None

  	# if no log facility, just print s
    def log(self, s):
        """
            Log to client's log facility or stdout
        """
        print s

# Reusable abstract class
# You don't like lame? implement your own encoder!
# TODO it should contain the name of the track to encode as a class member
class Encoder(UserDict):
    """
        Main Encoder interface
        Other encoders will be its childs
    """
	# A basic encoder only needs bitrate and samplerate, everything else is optional
    def __init__(self):
        UserDict.__init__(self)
        self["bitrate"] = 128
        self["samplerate"] = 44100
        self.format = "none"
        self.pid = None
        self._dump = None

    def encode(self):
        pass

    def stop_encode(self):
        """
            Stop encoding
        """
        if self.pid:
            subprocess.call("kill -9 %d" % self.pid, shell=True)
            self.pid = None

    def get_dump(self):
        return self._dump

    # This might not be enough at some point with slow computers,
    # we need to get a way to know the size of the file and exit
    # until enough data's been written by the encoder
    def sync(self):
        """
            Give some time to the encoder to write necessary data before streaming
        """
        while True:
            if self.pid:
                time.sleep(1.0)
                break
            time.sleep(0.01)
    
    # TODO this should be private, it'll be possible when encode gets implemented here
    def create_dump(self):
        if self._dump:
            self._dump.close()
        self._dump = tempfile.NamedTemporaryFile(suffix=".%s" \
                % self.format)

# TODO only support MP3 and OGG, other formats might be supported in the future
# (actually just need someone to reimplement the LameEncoder() provided)
# Reusable class
class Icast(UserDict):
    """
        Streaming to an icecast/shoutcast server
    """
    class LameEncoder(Encoder):
        """
           Lame MP3 Encoder
        """
        def __init__(self):
            Encoder.__init__(self)
            self.format = "mp3"

        # THIS IS HARD ON CPU SINCE IT NEEDS TO REENCODE THE SONG BEING CURRENTLY PLAYED BY THE CLIENT ON THE FLY,
        # WELL, SLOW ONLY WHILE IT'S ENCODING
        # TODO Improve so it gets deeepeer on the cpu
        # TODO this can be placed in the Encoder abstract class, making it easier to implement new encoders
        def encode(self, input):
            # The dump has to be created everytime we encode
            self.create_dump()
            #p1 = subprocess.Popen('arecord -t wav -f cd -D copy >> /dev/stdout', stdout=subprocess.PIPE, shell=True)
            p1 = subprocess.Popen('cat "%s"' % input, stdout=subprocess.PIPE, shell=True)
            p2 = subprocess.Popen('lame --mp3input --cbr -b %d --resample %d - - > %s' \
                    % (self["bitrate"], self["samplerate"],  self.get_dump().name), \
                    stdin=p1.stdout, stdout=subprocess.PIPE, shell=True)
            self.pid = p1.pid
            p1.wait()
            self.pid = None

    # Don't use me, I'm buggy :)
    class OggEncoder(Encoder):
        """
            Oggenc interface
        """
        def __init__(self):
            Encoder.__init__(self)
            self.format = "ogg"
        
        def encode(self, input):
            self.create_dump()
            # I don't know how to reencode an ogg track on the fly... HELP!
            p1 = subprocess.Popen('oggdec "%s" >> /dev/stdout' \
                    % input, stdout=subprocess.PIPE, shell=True)
            p2 = subprocess.Popen('oggenc --managed -b %d --resample %d - > "%s"' \
                    % (self["bitrate"], self["samplerate"], self.get_dump().name), \
                    stdin=p1.stdout, stdout=subprocess.PIPE, shell=True)
            self.pid = p1.pid
            p1.wait()
            self.pid = None

    def __init__(self):
        UserDict.__init__(self)
        self._server = shout.Shout()
        self.client = Client(None)
        self.connected = False
        self.current_track = None
        self.reencode = False
        self.broadcasting_live = False
    	# Default encoder
        self.encoder = self.LameEncoder()
        #self.encoder = self.OggEncoder()
        self.keep_alive = True
        self["host"] = "localhost"
        self["port"] = 8000
        self["pass"] = "hackme"
        self["user"] = "source"
        self["mount"] = "/icast"
        self["protocol"] = "http"
        self["format"] = "mp3"
        self["name"] = "Icast Stream"
        self["genre"] = "none"
        self["url"] = "none"
        self["description"] = "none"
        self["bitrate"] = 128
        self["samplerate"] = 44100

    def connect(self):
        """
            Connect to server
        """
        if not self.connected:
            self.client.log("Connecting to server %s:%d" % (self["host"], self["port"]))
            self._server.host = self["host"]
            self._server.port = self["port"]
            self._server.user = self["user"]
            self._server.password = self["pass"]
            self._server.mount = self["mount"]
            self._server.protocol = self["protocol"]
            self._server.format = self["format"]
#            if self.broadcasting_live:
#                self._server.nonblocking = True
            self._server.name = self["name"]
            self._server.genre = self["genre"]
            self._server.url = self["url"]
            self._server.description = self["description"]
            self._server.audio_info = {"bitrate":str(self["bitrate"]), 
                    "samplerate":str(self["samplerate"])}
            # We still don't support ogg reencoding
            if self.reencode and self["format"] == "ogg":
                self.reencode = False
            try:
                self._server.open()
            except shout.ShoutException:
                self.client.log("Failed connecting to server")
                raise Exception
            self.connected = True
            self.client.log("Connection with server established")

    def disconnect(self):
        """
            Disconnect from server
        """
        if self.connected:
            self._server.close()
            self.connected = False
            self.client.close()
            if self.reencode:
                self.encoder.stop_encode()
            self.client.log("Disconnected from server")

    def stop(self):
        self.disconnect()
        self.keep_alive = False
    
    @common.threaded
    def _reencode_track(self):
        """
            Call encoder to reencode current track
        """
        # We make sure we are the only thread reencoding
        self.encoder.stop_encode()
        self.encoder["bitrate"] = self["bitrate"]
        self.encoder["samplerate"] = self["samplerate"]
        self.client.log("Reencoding track with bittrate %skbps and samplerate %shz"
                % (self.encoder["bitrate"], self.encoder["samplerate"]))
        self.encoder.encode(self.client.get_current_track())
        self.client.log("Done reencoding '%s'" \
                % self.current_track)
  
    # data requested from the file can be changed to increase time to wait for the server, this is the key to make 
	# reencoding go easier on the cpu by pausing the encode_stream() thread
    # icecast/shoutcast only support mp3 and ogg, (maybe mp4 too), that should be implemented (but the encoder can 
	# accept any format, it won't be affected since the dump file it returns should be one of the two supported formats =)
	
    def _send_data(self, f):
        """
            Send data to the server
        """
        buf = f.read(4096*4)
        if len(buf) == 0:
            self.client.log("Done sending '%s'" \
                    % self.current_track)
            return False
        try:
            self._server.send(buf)
        except shout.ShoutException:
            self.disconnect()
            return False
        delay = self._server.delay()/1000.0
        if delay > 0: time.sleep(delay)
        return True

    def _change_data(self, f=None):
        """
            Update info for data to be streamed
        """
        self.current_track = self.client.get_next_track()
        if not self.current_track: return None
        # TODO this should appear in clients window too
        self.client.log("Streaming song %s" % self.current_track)
        try:
            self._server.set_metadata({
		        "song": str(self.client.get_track_metadata())
			})
        except shout.ShoutException:
            self.client.log("Sending metadata to the server not available" )
        if f:
            if self.broadcasting_live:
                return f
            else:
                f.close()
        if self.reencode:
            self._reencode_track()
            # give the encoder time to write necessary data before streaming
            self.encoder.sync()
            return self.encoder.get_dump()
        return open(self.current_track)

    def _close_stream(self, f = None):
        """
            Close the current stream
        """
        # TODO this should appear in clients window too
        self.client.log("Closing the stream...")
        if f:
            f.close()
        self.disconnect()

    # TODO method should provide a BETTER way to exit from the loop
    def icast_main(self):
        """
            Main Icast method, loops forever
        """
        connect_retry = False
        retry_delay = 3 #seconds
        f = None
        done = False
        sleep_delay = 0.01
        while self.keep_alive:
            if self.client.is_playing() \
                    and not done:
                # Connect if needed
                if not self.connected:
                    if connect_retry:
                        # Something happened, maybe just give the server some time to settle down
                        time.sleep(retry_delay)
                        connect_retry = False
                    try:
                        self.connect()
                    except Exception:
                        self.client.log("Connection retry count: 1")
                        connect_retry = True
                    else:
                        f = self._change_data()
                # Send data (and only do that! don't close the file!)
                elif not self._send_data(f):
                    # If done sending data, request another track
                    f = self._change_data(f)
                    if not f:
                        self.client.log("Finished sending data")
                        self._close_stream()
                        done = True
    		# if not playing
            if done and not self.client.is_playing():
                done = False
            elif self.connected and False:
                # if client stopped, disconnect from server
                # TODO how can we avoid this? we are suppose to only disconnect, not check if it's playing or not
                if not self.client.is_playing():
                    self._close_stream(f)
            else:
                time.sleep(sleep_delay)
 
# ----------------------------------------------------------------------------------------------------------------#
# Plugin code starts here. START HERE IF YOU NEED TO CHANGE ANY OPTION

PLUGIN_NAME = _("Icast Streamer")
PLUGIN_AUTHORS = ['Edgar Merino <donvodka at gmail dot com>']
PLUGIN_VERSION = "0.5.4"
__version = PLUGIN_VERSION
PLUGIN_DESCRIPTION = _(r"""Stream to an icecast/shoutcast server""")
PLUGIN_ENABLED = False
button = gtk.Button()
# TODO change with proper icon, is this really needed?
PLUGIN_ICON = button.render_icon('gnome-dev-usb', gtk.ICON_SIZE_MENU)
button.destroy()

PLUGIN = None

class ExaileClient(Client):
    """
        Exaile interface to the Icast Client
    """
    def __init__(self, app):
        Client.__init__(self, app)

    def is_playing(self):
  	    return self.app.player.is_playing()
  
    def get_current_track(self):
        return self.current_track.io_loc

    def get_next_track(self):
        if not self.current_track:
            self.current_track = self.app.player.current
        else:
            self.current_track = self.app.tracks.get_next_track(self.current_track)
        if not self.current_track: return None
        return self.current_track.io_loc

    def get_track_metadata(self):
        return "%s - %s" \
                % (self.current_track.get_artist(), \
                self.current_track.get_title())

    def log(self, s):
        xlmisc.log("IcastPlugin: %s" % s)

class IcastPlugin(Icast):
    """
        Icast Plugin for Exaile
    """
    def __init__(self):
        Icast.__init__(self)
        self.client = ExaileClient(APP)

        # IF YOU NEED TO CHANGE ANY OPTION TO CONNECT TO THE ICECAST/SHOUTCAST
        # SERVER THIS IS THE PLACE TO DO IT AT LEAST THE "PASS" KEY MUST BE CHANGED

        #self["host"] = "localhost"
        #self["port"] = 8000
        #self["user"] = "source"
        self["pass"] = "hackme"
        #self["mount"] = "/icast"

        # PROTOCOLS SUPPORTED:
        # "http" : icecast 2
        # "xaudiocast" : icecast
        # "icy" : shoutcast
        #self["protocol"] = "http"
    
        # FORMATS SUPPORTED:
        # "mp3", "ogg"
        #self["format"] = "mp3"

        #self["name"] = "Icast Station"
        #self["genre"] = "none"
        #self["url"] = "none"
        #self["description"] = "none"

        # ENABLE REENCODING, MP3 ONLY
        self.reencode = True
        self.encoder = self.LameEncoder()
        # THERE'S CURRENTLY NO SUPPORT TO SEND LIVE AUDIO FROM SOUNDCARD OUTPUT, DON'T CHANGE THIS
        # self.broadcasting_live = False 
        # if you want to set an encoder different than LameEncoder, this is the place to do it
        # this can be changed anywhere, we support live reencoding
        #self["bitrate"] = 128 # in kbps
        #self["samplerate"] = 44000 # in Hz

    @common.threaded
    def main(self):
        self.icast_main()

FIELDS = {
    'host': 'localhost',
    'port': 8000,
    'pass': 'hackme',
    'user': 'source',
    'mount': '/icast',
    'protocol': 'protocol',
    'format': 'mp3',
    'name': 'Icast Stream',
    'genre': 'none',
    'url': 'http://localhost:8000',
    'bitrate': 128,
    'samplerate': 44100
}

INTS = (8000, 128, 44100)

# TODO include configuration for the encoder too
def configure():
    """
        Allows for configuration of the icastplugin
    """
    global PLUGIN
    xml = gtk.glade.xml_new_from_buffer(XML_STRING, len(XML_STRING))
    dialog = xml.get_widget('ConfigDialog')

    for k, v in FIELDS.iteritems():
        func = 'get_str'
        if v in INTS: func = 'get_int'
        func = getattr(APP.settings, func)

        value = func(k, default=v, plugin=plugins.name(__file__))
        widget = xml.get_widget(k)

        if isinstance(widget, gtk.ComboBox):
            i = 0; selected = 0
            for item in widget.get_model():
                if item == v: selected = i
                i += 1

            widget.set_active(selected)
        else:
            widget.set_text(str(value))

    result = dialog.run()
    dialog.hide()

    if result == gtk.RESPONSE_OK:
        for k, v in FIELDS.iteritems():
            func = 'set_str'
            if v in INTS: func = 'set_int'
            func = getattr(APP.settings, func)

            widget = xml.get_widget(k)
            if isinstance(widget, gtk.ComboBox):
                value = widget.get_active_text()
            else:
                value = widget.get_text()

            if v in INTS: value = int(value)
            func(k, value, plugin=plugins.name(__file__))

        if PLUGIN:
            PLUGIN.stop()
            PLUGIN = setup_plugin()
            PLUGIN.main()

def setup_plugin():
    """
        Sets up the plugin with the values in the settings
    """
    plugin = IcastPlugin()

    for k, v in FIELDS.iteritems():
        func = 'get_str'
        if v in INTS: func = 'get_int'
        func = getattr(APP.settings, func)

        value = func(k, default=v, plugin=plugins.name(__file__))
        plugin[k] = value

    return plugin

def initialize():
    global PLUGIN

    exaile = APP
    if not SHOUT_AVAIL:
        # Long string... Mathias said this won't work for gettext if not done like this
        common.error(APP.window, _("Shout library could not be loaded. You need libshout 2 and shout python bindings. Streaming will not be available"))
        return False
    PLUGIN = setup_plugin()
    PLUGIN.main()

    return True

def destroy():
    global PLUGIN

    if PLUGIN:
        PLUGIN.stop()

    PLUGIN = None

XML_STRING = None
def load_data(zip):
    global XML_STRING
    XML_STRING = zip.get_data('gui.glade')



import mutagen.wavpack

def get_tag(wv, tag):
    try:
        return unicode(wv[tag][0])
    except:
        return ""

def fill_tag_from_path(tr):
    """
        Reads all tags from the file
    """
    f = mutagen.wavpack.WavPack(tr.io_loc)

    tr.length = int(f.info.length)

    tr.artist = get_tag(f, "artist")
    tr.album = get_tag(f, "album")
    tr.title = get_tag(f, "title")
    tr.genre = get_tag(f, "genre")
    tr.year = get_tag(f, "year")
    track = get_tag(f, "track")
    
    if track.find('/') > -1:
        (tr.track, tr.disc_id) = get_tag(f, "track").split('/')
    else:
        tr.track = get_tag(f, "track")
        tr.disc_id = ""
    
def write_tag(tr):
    """
        Writes all tags to the file
    """
    f = mutagen.wavpack.WavPack(tr.io_loc)

    f['artist'] = tr.artist
    f['album'] = tr.album
    f['title'] = tr.title
    f['genre'] = tr.genre
    f['year'] = tr.year

    if tr.disc_id:
        f['track'] = str(tr.track) + '/' + tr.disc_id
    else:
        f['track'] = str(tr.track)
    f.save()



# Global path variables and methods
# Copyright (c) 2007 Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os
from xl import logger

home = os.path.expanduser(u'~')
_configdir = os.path.join(home, '.exaile')
_cachedir = os.path.join(_configdir, 'cache')

# These are set in init
_datadir = None
localedir = None
firstrun = None

def init(basedir, installed):
    global _datadir, localedir, firstrun

    if installed:
        prefix = os.path.normpath(os.path.join(basedir, '../..'))
        _datadir = os.path.join(prefix, 'share', 'exaile')
        localedir = os.path.join(prefix, 'share', 'locale')
    else:
        _datadir = basedir
        localedir = os.path.join(basedir, 'po')

    firstrun = not os.path.isdir(_configdir)
    if firstrun:
        try:
            os.mkdir(_configdir)
        except:
            # FIXME: Die?
            logger.log("Could not create settings directory")
        try:
            os.mkdir(_cachedir)
        except:
            # FIXME: Die?
            logger.log("Could not create cache directory")

def set_configdir(dir):
    global _configdir, _cachedir
    _configdir = dir
    _cachedir = os.path.join(_configdir, 'cache')

def get_cache(*path_elems):
    return os.path.join(_cachedir, *path_elems)

def get_config(*path_elems):
    if path_elems[0] == 'cache':
        logger.log_stack("WARNING: get_config called for 'cache', "
            "use get_cache instead")
    return os.path.join(_configdir, *path_elems)

def get_data(*path_elems):
    return os.path.join(_datadir, *path_elems)



#!/usr/bin/env python

# web_controller - Simple Web server to control Exaile
# Copyright (c) 2007 Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Simple Web server to control Exaile.

Usage: web_controller [port [exaile_executable]]

Note that this server is not secured at all other than by a strict limitation of
the available commands.  Anyone with access to the server will still be able to
control playback.
"""

from subprocess import *
from BaseHTTPServer import *

class ExaileHttpHandler(BaseHTTPRequestHandler):
    allowed_commands = ['prev', 'play', 'play-pause', 'stop', 'next', 'query']

    def do_GET(self):
        self.send_response(200)
        self.send_header('Content-type', 'text/html')
        self.end_headers()

        write = self.wfile.write
        write('<html><head><title>Exaile</title><body>')

        if self.path.startswith('/'):
            command = self.path[1:]
        else:
            command = None

        if command and command in self.allowed_commands:
            line = [self.server.exaile, '--' + command]
            print 'Running', line
            output = Popen(line, stdout=PIPE).communicate()[0]
            write('<pre>' + output + '</pre>')

        write('<ul>')
        for cmd in self.allowed_commands:
            write('<li><a href="' + cmd + '">' + cmd + '</a></li>')
        write('</ul>')

        write('</body></html>')

def run(port, exaile):
    server = None
    try:
        server = HTTPServer(('', port), ExaileHttpHandler)
        server.exaile = exaile
        server.serve_forever()
    except KeyboardInterrupt:
        pass
    finally:
        if server:
            server.socket.close()

if __name__ == '__main__':
    from sys import argv
    argc = len(argv)
    port = 8080
    exaile = 'exaile'
    if argc > 1:
        port = int(argv[1])
        if argc > 2:
            exaile = argv[2]
    run(port, exaile)

# vi: et ts=4 sts=4 sw=4 tw=80



# Filter Widget/Dialog
# Copyright (c) 2006 Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Generic widget and dialog for filtering items.

They resemble the configuration dialogs of Evolution's mail filters
and Rhythmbox's automatic playlists.
"""

from gettext import gettext as _

import gtk

class FilterDialog(gtk.Dialog):
    """Dialog to filter a list of items.

    Consists of a FilterWidget and an Add button.
    """

    def __init__(self, title, criteria):
        """Create a filter dialog.

        Parameters:
        - title: title of the dialog window
        - criteria: possible criteria; see FilterWindow
        """

        gtk.Dialog.__init__(self, title, buttons=(
            gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT,
            gtk.STOCK_OK, gtk.RESPONSE_ACCEPT))

        top = gtk.HBox()
        top.set_border_width(5)
        top.set_spacing(5)

        top.pack_start(gtk.Label(_("Name:")), False)
        self.name_entry = gtk.Entry()
        top.pack_start(self.name_entry)
        self.vbox.pack_start(top)
        top.show_all()

        self.filter = f = FilterWidget(criteria)
        f.add_row()
        f.set_border_width(5)
        self.vbox.pack_start(f)
        f.show()

        bottom = gtk.HBox()
        bottom.set_border_width(5)
        self.match_any = gtk.CheckButton(_('Match any of the criteria'))
        bottom.pack_start(self.match_any)

        btn = gtk.Button()
        btn.connect('clicked', lambda *x: self.filter.add_row())
        image = gtk.Image()
        image.set_from_stock(gtk.STOCK_ADD, gtk.ICON_SIZE_BUTTON)
        btn.add(image)
        align = gtk.Alignment(xalign=1)
        align.add(btn)
        bottom.pack_end(align)
        self.vbox.pack_start(bottom)

        # add the limit checkbox, spinner
        limit_area = gtk.HBox()
        limit_area.set_border_width(5)
        self.lim_check = gtk.CheckButton(_("Limit to: "))
        limit_area.pack_start(self.lim_check, False)

        self.lim_spin = gtk.SpinButton(gtk.Adjustment(1, 0, 1000000000, 1))
        self.lim_spin.set_sensitive(False)

        self.lim_check.connect('toggled', lambda b:
            self.lim_spin.set_sensitive(self.lim_check.get_active()))

        limit_area.pack_start(self.lim_spin, False)
        limit_area.pack_start(gtk.Label(_(" songs")), False)
        self.vbox.pack_start(limit_area)
        limit_area.show_all()

        bottom.show_all()
        align.show_all()

    def set_limit(self, limit):
        """
            Sets the limit for the number of items that should be returned
        """
        if limit > 0:
            self.lim_check.set_active(True)
            self.lim_spin.set_value(limit)
        else:
            self.lim_check.set_active(False)

    def get_limit(self):
        """
            Get the limit value
        """
        if self.lim_check.get_active():
            return int(self.lim_spin.get_value())
        else:
            return 0

    def get_match_any(self):
        """
            Returns true if this dialog should match any of the criteria
        """
        return self.match_any.get_active()

    def set_match_any(self, any):
        """
            Sets whether this dialog should match any of the criteria or not
        """
        self.match_any.set_active(any)

    def get_name(self):
        """
            Returns the text in the name_entry
        """
        return unicode(self.name_entry.get_text(), 'utf-8')

    def set_name(self, name):
        """
            Sets the text in the name_entry
        """
        self.name_entry.set_text(name)

    def get_result(self):
        """Return the user's input as a list of filter criteria.
        
        See FilterWidget.get_result.
        """
        # TODO: display error message
        return self.filter.get_result()

    def get_state(self):
        """Return the filter state.
        
        See FilterWidget.get_state.
        """
        return self.filter.get_state()

    def set_state(self, state):
        """Set the filter state.
        
        See FilterWidget.set_state.
        """
        self.filter.set_state(state)

class FilterWidget(gtk.Table):
    """Widget to filter a list of items.

    This widget only includes the criteria selector (Criterion widgets)
    and their Remove buttons; it does not include an Add button.

    Attributes:
    - criteria: see example
    - rows: list of (criterion, remove_btn, remove_btn_handler_id)
    - n: number of rows

    Example of criteria:

        [
            # name
            (N_('Year'), [
                # name - field class/factory -  result generator
                (N_('is'), (EntryField, lambda x: 'Year=%d' % int(x))),
                (N_('is between'), (
                    lambda x: EntryLabelEntryField(x, _('and')),
                    lambda x, y:
                        'Year BETWEEN %s AND %s' % (int(x), int(y)))),
                (N_('is this year'), (NothingField,
                    lambda: time.localtime()[0])),
            ]),
        ]

    The field class is a class following this interface:

        class Field(gtk.Widget):
            def __init__(self, result_generator):
                pass
            def get_result(self):
                return object
            def get_state(self):
                return object
            def set_state(self, state):
                pass

    Although not required, these methods should use unicode instead of
    str objects when dealing with strings.
    """

    def __init__(self, criteria):
        """Create a filter widget.

        Parameter:
        - criteria: see FilterWidget
        """

        gtk.Table.__init__(self)
        self.set_col_spacings(10)
        self.set_row_spacings(2)
        self.criteria = criteria
        self.rows = []
        self.n = 0

    def get_result(self):
        """Return a list of results produced by the criterion objects.
        """
        return [f[0].get_result() for f in self.rows]

    def add_row(self):
        """Add a new criteria row."""

        criterion = Criterion(self.criteria)
        criterion.show()

        remove_btn = gtk.Button()
        image = gtk.Image()
        image.set_from_stock(gtk.STOCK_REMOVE, gtk.ICON_SIZE_BUTTON)
        remove_btn.add(image)
        remove_btn_handler_id = remove_btn.connect(
            'clicked', self.__remove_clicked, self.n)
        remove_btn.show_all()

        self.attach(criterion, 0, 1, self.n, self.n + 1,
            gtk.EXPAND | gtk.FILL, gtk.SHRINK)
        self.attach(remove_btn, 1, 2, self.n, self.n + 1,
            gtk.FILL, gtk.SHRINK)

        self.rows.append((criterion, remove_btn, remove_btn_handler_id))
        self.n += 1

    def remove_row(self, row):
        """Remove a criteria row."""
        rows = self.rows
        for iRow in xrange(row, len(rows)):
            crit, btn, handler = rows[iRow]
            self.remove(crit)
            self.remove(btn)
            btn.disconnect(handler)
            if iRow != row:  # shift up
                self.attach(crit, 0, 1, iRow - 1, iRow,
                    gtk.EXPAND | gtk.FILL, gtk.SHRINK)
                self.attach(btn, 1, 2, iRow - 1, iRow,
                    gtk.FILL, gtk.SHRINK)
                handler = btn.connect(
                    'clicked', self.__remove_clicked, iRow - 1)
                rows[iRow - 1] = crit, btn, handler
        self.n -= 1
        del rows[self.n]
        if self.n:
            self.resize(self.n, 2)

    def __remove_clicked(self, widget, data):
        self.remove_row(data)

    def get_state(self):
        """Return the filter state.

        See set_state for the state format.
        """
        state = []
        for row in self.rows:
            state.append(row[0].get_state())
            state[-1][0].reverse() # reverse so it reads more nicely
        return state

    def set_state(self, state):
        """Set the filter state.

        Format:

            [
                ( [criterion1_1, criterion1_2, ...], filter1 ),
                ( [criterion2_1, criterion2_2, ...], filter2 ),
                ...
            ]
        """
        n_present = len(self.rows)
        n_required = len(state)
        for i in xrange(n_present, n_required):
            self.add_row()
        for i in xrange(n_present, n_required, -1):
            self.remove_row(i - 1) # i is one less than n
        for i, cstate in enumerate(state):
            cstate[0].reverse() # reverse so it becomes a stack
            self.rows[i][0].set_state(cstate)

class Criterion(gtk.HBox):
    """Widget representing one filter criterion.

    It contains either:
    - one combo box and another criterion object, or
    - a field object.
    """

    def __init__(self, subcriteria):
        """Create a criterion object.

        Parameters:
        - subcriteria: a list of possible subcriteria;
          see criteria in FilterWidget
        """
        gtk.HBox.__init__(self, spacing=5)
        if isinstance(subcriteria, tuple):
            field_class, result_generator = subcriteria
            self.child = field_class(result_generator)
        else:
            self.combo = combo = gtk.combo_box_new_text()
            self.subcriteria = subcriteria
            for subc in subcriteria:
                combo.append_text(_(subc[0]))
            combo.set_active(0)
            combo.connect('changed', self._combo_changed)
            combo.show()
            self.pack_start(combo, False)
            self.child = Criterion(subcriteria[0][1])
        self.child.show()
        self.pack_start(self.child)

    def _combo_changed(self, widget):
        """Called when the combo box changes its value."""
        state = self.child.get_state() 
        self.remove(self.child)
        self.child = Criterion(self.subcriteria[self.combo.get_active()][1])
        if state: self.child.set_state(state)
        self.pack_start(self.child)
        self.child.show()

    def get_result(self):
        """Return the result from the field object."""
        return self.child.get_result()

    def get_state(self):
        """Return the criterion state.

        See set_state for the state format.
        """
        state = self.child.get_state()
        if isinstance(self.child, Criterion):
            state[0].append(self.subcriteria[self.combo.get_active()][0])
        else:
            state = ([], state)
        return state

    def set_state(self, state):
        """Set the criterion state.

        Format:

            ([..., grandchild_state, child_state, self_state], filter)

        Note the reverse order of the list. This is to give the
        impression of it being a stack responding to pop().
        """
        if isinstance(self.child, Criterion):
            text = state[0].pop()
            for i, subc in enumerate(self.subcriteria):
                if subc[0] == text:
                    self.combo.set_active(i)
                    break
            self.child.set_state(state)
        else:
            if len(state) > 1: 
                self.child.set_state(state[1])

# Sample fields

class MultiEntryField(gtk.HBox):
    """Helper field that can be subclassed to get fields with multiple
       GtkEntry widgets and multiple labels."""
    def __init__(self, result_generator, n=2, labels=None, widths=None):
        gtk.HBox.__init__(self, spacing=5)
        self.generate_result = result_generator
        self.entries = []
        for iEntry in xrange(n):
            entry = gtk.Entry()
            if iEntry < len(widths):
                w = widths[iEntry]
                if w is not None:
                    entry.set_size_request(w, -1)
            self.entries.append(entry)
            if iEntry < len(labels):
                txt = labels[iEntry]
                if txt is not None:
                    l = gtk.Label(txt)
                    self.pack_start(l, False)
                    l.show()
            self.pack_start(entry)
            entry.show()
        if n < len(labels):
            txt = labels[n]
            if txt is not None:
                l = gtk.Label(txt)
                self.pack_start(l, False)
                l.show()
    def get_result(self):
        return self.generate_result(*self.get_state())
    def get_state(self):
        return [unicode(e.get_text(), 'utf-8') for e in self.entries]
    def set_state(self, state):
        for i, e in enumerate(self.entries):
            if len(state) > i: e.set_text(unicode(state[i]))

class EntryField(gtk.Entry):
    def __init__(self, result_generator):
        gtk.Entry.__init__(self)
        self.generate_result = result_generator
    def get_result(self):
        return self.generate_result(unicode(self.get_text(), 'utf-8'))
    def get_state(self):
        return self.get_text()
    def set_state(self, state):
        if type(state) == list or type(state) == tuple:
            state = state[0]
        self.set_text(unicode(state))

class EntryLabelEntryField(MultiEntryField):
    def __init__(self, result_generator, label):
        MultiEntryField.__init__(self, result_generator, n=2,
            labels=(None, label, None),
            widths=(50, 50))

class SpinLabelField(gtk.HBox):
    def __init__(self, result_generator, label='', top=99999, lower=-99999):
        gtk.HBox.__init__(self, spacing=5)
        self.generate_result = result_generator
        self.spin = gtk.SpinButton(gtk.Adjustment(0, lower, top, 1, 0, 0))
        self.pack_start(self.spin)
        self.pack_start(gtk.Label(label))
        self.show_all()
    def get_state(self):
        return self.spin.get_value()
    def set_state(self, state):
        if type(state) == list or type(state) == tuple:
            state = state[0]
        try:
            self.spin.set_value(int(state))
        except ValueError:
            pass

    def get_result(self):
        return self.generate_result(self.spin.get_value())

class SpinButtonAndComboField(gtk.HBox):
    def __init__(self, result_generator, items=()):
        gtk.HBox.__init__(self, spacing=5)
        self.generate_result = result_generator

        adjustment = gtk.Adjustment(0, 0, 99999, 1, 0, 0)
        self.entry = gtk.SpinButton(adjustment=adjustment)
        self.pack_start(self.entry)

        self.combo = gtk.combo_box_new_text()
        for item in items:
            self.combo.append_text(item)
        self.combo.set_active(0)
        self.pack_start(self.combo)
        self.show_all()

    def get_result(self):
        return self.generate_result(*self.get_state())

    def set_state(self, state):
        if not isinstance(state, (tuple, list)):
            return

        print state

        # TODO: Check length.
        try:
            self.entry.set_value(int(state[0]))
        except ValueError:
            pass
        count = 0
        model = self.combo.get_model()
        iter = model.get_iter_first()
        while True:
            text = model.get_value(iter, 0)
            if text == state[1]:
                self.combo.set_active(count)

            count += 1
            iter = model.iter_next(iter)
            if not iter: break

    def get_state(self):
        return [self.entry.get_value(), 
            unicode(self.combo.get_active_text(), 'utf-8')]



#!/usr/bin/env python
"""This file is automatically generated by generate_version_info
It uses the current working tree to determine the revision.
So don't edit it. :)
"""

version_info = {'branch_nick': u'main',
 'build_date': '2008-08-22 21:25:50 -0500',
 'clean': None,
 'date': '2008-08-02 01:15:31 +0200',
 'revision_id': 'info@noctus.net-20080801231531-6wufpnd7lzhhxrv6',
 'revno': 1650}

revisions = {}

file_revisions = {}



if __name__ == '__main__':
    print 'revision: %(revno)d' % version_info
    print 'nick: %(branch_nick)s' % version_info
    print 'revision id: %(revision_id)s' % version_info



# Copyright (C) 2006 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import md5, os, re, threading, time, urllib
from lib import feedparser
from gettext import gettext as _
import gobject, gtk
from xl import common, xlmisc, library, media
from xl.gui import playlist as trackslist
import xl.path

class CustomWrapper(object):
    """
        Wraps a custom radio station
    """
    def __init__(self, name):
        """
            initializes the wrapper
        """
        self.name = name

    def __str__(self):
        """
            Returns the name
        """
        return self.name

class PodcastWrapper(object):
    """
        Wraps a podcast
    """
    def __init__(self, name, path):
        """
            Initializes the wrapper
        """
        self.name = name
        self.path = path

    def __str__(self):
        """
            Returns the name
        """
        return self.name

class PodcastQueueThread(threading.Thread):
    """
        Downloads podcasts in the queue one by one
    """
    def __init__(self, transfer_queue, panel):
        """ 
            Initializes the transfer
        """
        threading.Thread.__init__(self)
        self.transfer_queue = transfer_queue
        self.panel = panel
        self.queue = transfer_queue.queue
        self.setDaemon(True)
        self.stopped = False

    def run(self):
        """
            Actually runs the thread
        """
        for song in self.queue: 
            (download_path, downloaded) = \
                self.panel.get_podcast_download_path(song.loc)
            if self.stopped: break
            hin = urllib.urlopen(song.loc)

            temp_path = "%s.part" % download_path
            hout = open(temp_path, "w+")

            count = 0
            while True:
                data = hin.read(1024)
                self.transfer_queue.downloaded_bytes += len(data)
                if count >= 10:
                    gobject.idle_add(self.transfer_queue.update_progress)
                    count = 0
                if not data: break
                hout.write(data)
                if self.stopped:
                    hout.close()
                    os.unlink(temp_path)
                    break

                count += 1

            hin.close()
            hout.close()

            if os.path.isfile(temp_path):
                os.rename(temp_path, download_path)
            if self.stopped: break

            self.transfer_queue.downloaded += 1
            gobject.idle_add(self.transfer_queue.update_progress)
            song.download_path = download_path
            temp = library.read_track(None, None, download_path)

            if temp:
                song.set_len(temp.duration)
                self.db = self.panel.exaile.db
                path_id = library.get_column_id(self.db, 'paths', 'name',
                    song.loc)
                podcast_path_id = library.get_column_id(self.db, 'paths', 
                    'name', song.podcast_path)

                self.panel.exaile.db.execute(
                    "UPDATE podcast_items SET length=? WHERE podcast_path=?"
                    " AND path=?", 
                    (temp.duration, podcast_path_id, path_id))
                self.panel.exaile.db.commit()

            gobject.idle_add(self.transfer_queue.update_song, song)
            xlmisc.log("Downloaded podcast %s" % song.loc)

        gobject.idle_add(self.transfer_queue.die)

class PodcastTransferQueue(gtk.VBox):
    """
        Represents the podcasts that should be downloaded
    """
    def __init__(self, panel):
        """
            Starts the transfer queue
        """
        gtk.VBox.__init__(self)
        self.panel = panel

        self.label = gtk.Label(_("Downloading Podcasts"))
        self.label.set_alignment(0, 0)

        self.pack_start(self.label)
        self.progress = gtk.ProgressBar()
        self.progress.set_text(_("Downloading..."))
        self.progress.set_size_request(-1, 24)
        
        vert = gtk.HBox()
        vert.set_spacing(3)
        vert.pack_start(self.progress, True, True)
        
        button = gtk.Button()
        image = gtk.Image()
        image.set_from_stock('gtk-stop', 
            gtk.ICON_SIZE_SMALL_TOOLBAR)
        button.set_image(image)
        button.set_size_request(32, 32)

        vert.pack_end(button, False, False)
        button.connect('clicked', self.stop)

        self.pack_start(vert)

        self.downloaded = 0
        self.downloaded_bytes = 0
        self.total = 0
        self.total_bytes = 0
        self.queue = library.TrackData()
        self.queue_thread = None
        panel.podcast_download_box.pack_start(self)
        self.show_all()

    def stop(self, widget):
        """
            Stops the download queue
        """
        if self.queue_thread:
            self.queue_thread.stopped = True

        self.label.set_label(_("Stopping..."))

    def update_song(self, song):
        """ 
            Updates song information in the display
        """
        nb = self.panel.exaile.playlists_nb
        for i in range(nb.get_n_pages()):
            page = nb.get_nth_page(i)
            if isinstance(page, trackslist.TracksListCtrl):
                for item in page.songs:
                    if item == song:
                        page.refresh_row(song)

    def update_progress(self):
        """ 
            Update the progress bar with the percent of downloaded items
        """
        if self.total_bytes:
            total = float(self.total_bytes)
            down = float(self.downloaded_bytes)
            percent = down / total
            self.progress.set_fraction(percent)

        self.label.set_label(_("%(downloaded)d of %(total)d downloaded") % 
            {
                'downloaded': self.downloaded,
                'total': self.total
            })

    def append(self, song):
        """
            Appends an item to the queue
        """
        self.queue.append(song)
        self.total += 1
        self.total_bytes += song.size
        self.update_progress()
        if not self.queue_thread:
            self.queue_thread = PodcastQueueThread(self, self.panel)
            self.queue_thread.start()

    def die(self):
        """
            Removes the download queue
        """
        self.hide()
        self.panel.podcast_download_box.remove(self)
        self.panel.podcast_queue = None

class LastFMWrapper(object):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name


class EmptyRadioDriver(object):
    """
        Empty Driver
    """
    def __init__(self):
        pass

class RadioGenre(object):
    def __init__(self, name, driver=None, extra=None):
        self.name = name
        self.extra = extra
        self.driver = driver

    def __str__(self):
        return self.name

class RadioDriver(object):
    pass

class RadioPanel(object):
    """
        This will be a pluggable radio panel.  Plugins like shoutcast and
        live365 will go here
    """
    name = 'pradio'
    def __init__(self, exaile):
        """
            Initializes the panel
        """
        self.exaile = exaile
        self.db = exaile.db
        self.xml = exaile.xml
        self.tree = self.xml.get_widget('radio_service_tree')
        icon = gtk.CellRendererPixbuf()
        text = gtk.CellRendererText()
        col = gtk.TreeViewColumn('radio')
        col.pack_start(icon, False)
        col.pack_start(text, True)
        col.set_attributes(icon, pixbuf=0)
        col.set_cell_data_func(text, self.cell_data_func)
        self.tree.append_column(col)
        self.podcasts = {}
        self.drivers = {}
        self.driver_names = {}
        self.__dragging = False

        self.model = gtk.TreeStore(gtk.gdk.Pixbuf, object)
        self.tree.set_model(self.model)

        self.open_folder = xlmisc.get_icon('gnome-fs-directory-accept')
        self.track = gtk.gdk.pixbuf_new_from_file(xl.path.get_data('images',
            'track.png'))
        self.folder = xlmisc.get_icon('gnome-fs-directory')
        self.refresh_image = xlmisc.get_icon('gtk-refresh')

        self.track = gtk.gdk.pixbuf_new_from_file(xl.path.get_data('images',
            'track.png'))
        self.custom = self.model.append(None, [self.open_folder, _("Saved Stations")])
        self.podcast = self.model.append(None, [self.open_folder, _("Podcasts")])

        # load all saved stations from the database
        rows = self.db.select("SELECT name FROM radio "
            "ORDER BY name")
        for row in rows:
            self.model.append(self.custom, [self.track, CustomWrapper(row[0])])

        # load podcasts
        rows = self.db.select("SELECT title, paths.name FROM podcasts,paths "
            "WHERE paths.id=podcasts.path")
        for row in rows:
            title = row[0]
            path = row[1]
            if not title: title = path
            self.model.append(self.podcast, [self.track, 
                PodcastWrapper(title, path)])

        self.tree.expand_row(self.model.get_path(self.custom), False)
        self.tree.expand_row(self.model.get_path(self.podcast), False)

        self.radio_root = self.model.append(None, [self.open_folder, _("Radio "
            "Streams")])

        self.drivers_expanded = {}
        self.load_nodes = {}
        self.tree.connect('row-expanded', self.on_row_expand)
        self.tree.connect('button-press-event', self.button_pressed)
        self.tree.connect('button-release-event', self.button_release)
        self.tree.connect('row-collapsed', self.on_collapsed)
        self.tree.connect('button-press-event', self.button_pressed)
        self.tree.connect('button-release-event', self.button_release)
        self.__dragging = False
        self.xml.get_widget('pradio_add_button').connect('clicked',
            self.on_add_station)
        self.xml.get_widget('pradio_remove_button').connect('clicked',
            self.remove_station)
        self.podcast_download_box = \
            self.xml.get_widget('ppodcast_download_box')
        self.podcast_queue = None
        self.setup_menus()

    def button_release(self, widget, event):
        """
            Called when a button is released
        """
        if event.button != 1 or self.__dragging: return True
        if event.state & (gtk.gdk.SHIFT_MASK|gtk.gdk.CONTROL_MASK):
            return True
        selection = self.tree.get_selection()
        x, y = event.get_coords()
        x = int(x); y = int(y)

        path = self.tree.get_path_at_pos(x, y)
        if not path: return False
        selection.unselect_all()
        selection.select_path(path[0])

    def button_pressed(self, widget, event):
        """
            Called when the user clicks on the tree
        """
        selection = self.tree.get_selection()
        selection.unselect_all()
        (x, y) = event.get_coords()
        x = int(x); y = int(y)
        if not self.tree.get_path_at_pos(x, y): return
        (path, col, x, y) = self.tree.get_path_at_pos(x, y)
        selection.select_path(path)
        model = self.model
        iter = model.get_iter(path)
        
        object = model.get_value(iter, 1)
        if event.button == 3:
            # if it's for podcasts
            if isinstance(object, PodcastWrapper) or \
                object == "Podcasts":
                self.podmenu.popup(None, None, None,
                    event.button, event.time)
                return

            if isinstance(object, CustomWrapper):
                self.cmenu.popup(None, None, None,
                    event.button, event.time)

            elif isinstance(object, RadioDriver) or isinstance(object,
                RadioGenre):
                if isinstance(object, RadioDriver):
                    driver = object
                else:
                    driver = object.driver
                self.setup_menus()
                if driver and hasattr(driver, 'get_menu'):
                    menu = driver.get_menu(object, self.menu)
                else:
                    menu = self.menu
                menu.popup(None, None, None, event.button, event.time)
                return True
            else:
                if object in ("Saved Stations", "Podcasts", "Shoutcast Stations"):
                    return
        elif event.type == gtk.gdk._2BUTTON_PRESS:
            if object == 'Last.fm Radio':
                self.tree.expand_row(path, False)                
            elif isinstance(object, CustomWrapper):
                self.open_station(object.name)
            elif isinstance(object, PodcastWrapper):
                self.open_podcast(object)
            elif isinstance(object, LastFMWrapper):
                self.open_lastfm(object)
            elif isinstance(object, RadioGenre):
                if object.driver:
                    if not hasattr(object.driver, 'no_new_page') or not \
                        object.driver.no_new_page:
                        tracks = trackslist.TracksListCtrl(self.exaile)
                        self.exaile.playlists_nb.append_page(tracks,
                            xlmisc.NotebookTab(self.exaile, str(object), tracks))
                        self.exaile.playlists_nb.set_current_page(
                            self.exaile.playlists_nb.get_n_pages() - 1)
                        self.exaile.tracks = tracks
                        object.driver.tracks = tracks
                    object.driver.load_genre(object)
            return True

    def cell_data_func(self, column, cell, model, iter, user_data=None):
        """
            Called when the tree needs a value for column 1
        """
        object = model.get_value(iter, 1)
        if isinstance(object, CustomWrapper):
            cell.set_property('text', str(object))
        else:
            cell.set_property('text', str(object))

    def add_driver(self, driver, name):
        """
            Adds a driver to the list of drivers
        """
        if not self.drivers.has_key(driver):
            driver.name = name
            self.driver_names[driver] = name
            node = self.model.append(self.radio_root, [self.folder, driver])

            self.load_nodes[driver] = self.model.append(node, 
                [self.refresh_image, _("Loading streams...")])
            self.drivers[driver] = node
            self.tree.expand_row(self.model.get_path(self.radio_root), False)
            if self.exaile.settings.get_boolean('row_expanded', plugin=name,
                default=False):
                self.tree.expand_row(self.model.get_path(node), False)

    def remove_driver(self, driver):
        """
            Removes a radio driver
        """
        if self.drivers.has_key(driver):
            self.model.remove(self.drivers[driver])
            del self.drivers[driver]

    def open_lastfm(self, object):
        """
            Opens and plays a Last.fm station
        """
        station = str(object)
        user = self.exaile.settings.get_str('lastfm/user', '')
        password = self.exaile.settings.get_crypted('lastfm/pass', '')

        if not user or not password:
            common.error(self.exaile.window, _("You need to have a Last.fm "
                "username and password set in your preferences."))
            return

        if station == "Neighbor Radio":
            url = "lastfm://user/%s/neighbours" % user
        else:
            url = "lastfm://user/%s/personal" % user

        tr = media.Track(url)
        tr.type = 'lastfm'
        tr.track = -1
        tr.title = station
        tr.album = _("%(user)s's Last.fm %(station)s") % \
            {
                'user': user,
                'station': station
            }


        self.exaile.playlist_manager.append_songs((tr,))

    def open_podcast(self, wrapper):
        """
            Opens a podcast
        """
        podcast_path_id = library.get_column_id(self.db, 'paths', 'name',
            wrapper.path)

        xlmisc.log("Opening podcast %s" % wrapper.name)
        row = self.db.read_one("podcasts", "description", "path=?", 
            (podcast_path_id, ))
        if not row: return

        desc = row[0]
        limit = self.exaile.settings.get_str('podcast_show_limit', '0')
        limiter = ""
        if int(limit) > 0: limiter = "LIMIT %s" % limit
        rows = self.db.select("SELECT paths.name, title, description, length, "
            "pub_date, size FROM podcast_items, paths WHERE podcast_path=? "
            "AND paths.id=podcast_items.path ORDER BY "
            "pub_date DESC %s" % limiter, 
            (podcast_path_id,))

        songs = library.TrackData()
        for row in rows:
            t = common.strdate_to_time(row[4])
            year = time.strftime("%x", time.localtime(t))
            info = ({
                'title': row[1],
                'artist': row[2],
                'album': desc.replace('\n', ' '),
                'loc': row[0],
                'year': row[4], 
                'length': row[3], 
            })

            song = media.Track()
            song.set_info(**info)
            song.type = 'podcast'
            song.size = row[5]
            song.year = song.year.replace(' 00:00:00', '')

            (download_path, downloaded) = \
                self.get_podcast_download_path(row[0])

            if not downloaded:
                song.download_path = ''
            else:
                song.download_path = download_path

            song.podcast_path = wrapper.path
            songs.append(song)
            self.podcasts[song.loc] = song

        self.exaile.new_page(str(wrapper), songs)

    def get_podcast(self, url):
        """
            Returns the podcast for the specified url
        """
        if self.podcasts.has_key(url):
            return self.podcasts[url]
        else: return None

    def add_podcast_to_queue(self, song):
        """
            Add to podcast transfer queue
        """
        if not self.podcast_queue:
            self.podcast_queue = PodcastTransferQueue(self)

        if not song.loc in self.podcast_queue.queue.paths:
            self.podcast_queue.append(song)

    def get_podcast_download_path(self, loc):
        """
            Gets the location of the downloaded pocast item
        """
        (path, ext) = os.path.splitext(loc)
        hash = md5.new(loc).hexdigest()
        savepath = self.exaile.settings.get_str('feed_download_location',
            os.path.expanduser('~/.exaile/podcasts')) 
        if not os.path.isdir(savepath):
            os.mkdir(savepath, 0777)

        file = os.path.join(savepath, hash + ext)
        if os.path.isfile(file): return file, True
        else: return file, False

    def cell_data_func(self, column, cell, model, iter, user_data=None):
        """
            Called when the tree needs a value for column 1
        """
        object = model.get_value(iter, 1)
        if isinstance(object, CustomWrapper):
            cell.set_property('text', str(object))
        else:
            cell.set_property('text', str(object))
            
    def on_collapsed(self, tree, iter, path):
        """
            Called when someone collapses a tree item
        """
        driver = self.model.get_value(iter, 1)
        selection = self.tree.get_selection()
        (model, iter) = selection.get_selected()
        self.model.set_value(iter, 0, self.folder)
        self.tree.queue_draw()
        self.exaile.settings.set_boolean('row_expanded', False,
            plugin=self.driver_names[driver])

    def on_row_expand(self, treeview, iter, path):
        """
            Called when the user clicks on a row to expand the stations under
        """
        driver = self.model.get_value(iter, 1)
        self.model.set_value(iter, 0, self.open_folder)
        self.tree.queue_draw()

        if not isinstance(driver, RadioDriver): return
        if self.drivers.has_key(driver) and not \
            self.drivers_expanded.has_key(driver):
            self.drivers_expanded[driver] = 1

            driver.load_streams(self.drivers[driver],
                self.load_nodes[driver]) 
        self.exaile.settings.set_boolean('row_expanded', True,
            plugin=self.driver_names[driver])

    def setup_menus(self):
        """
            Create the two different popup menus associated with this tree.
            There are two menus, one for saved stations, and one for
            shoutcast stations
        """
        self.menu = xlmisc.Menu()
        rel = self.menu.append(_("Refresh"), lambda e, f:
            self.refresh_streams(), 'gtk-refresh')

        # custom playlist menu
        self.cmenu = xlmisc.Menu()
        self.add = self.cmenu.append(_("Add Stream to Station"), 
            self.add_url_to_station)
        self.delete = self.cmenu.append(_("Delete this Station"),
            self.remove_station)

        self.podmenu = xlmisc.Menu()
        self.podmenu.append(_("Add Feed"), self.on_add_podcast)
        self.podmenu.append(_("Refresh Feed"), self.refresh_feed)
        self.podmenu.append(_("Delete Feed"), self.delete_podcast)

    def refresh_streams(self):
        """
            Refreshes the streams for the currently selected driver
        """
        selection = self.tree.get_selection()
        (model, iter) = selection.get_selected()
        object = self.model.get_value(iter, 1)
        if isinstance(object, RadioDriver):
            driver = object
            self.drivers_expanded[driver] = 1
            self.clean_node(self.drivers[driver])
            self.load_nodes[driver] = self.model.append(iter, 
                [self.refresh_image, _("Loading streams...")])
            driver.load_streams(self.drivers[driver], 
                self.load_nodes[driver], False)
            self.tree.expand_row(self.model.get_path(iter), False)
        elif isinstance(object, RadioGenre):
            if object.driver:
                tracks = trackslist.TracksListCtrl(self.exaile)
                self.exaile.playlists_nb.append_page(tracks,
                    xlmisc.NotebookTab(self.exaile, str(object), tracks))
                self.exaile.playlists_nb.set_current_page(
                    self.exaile.playlists_nb.get_n_pages() - 1)
                self.exaile.tracks = tracks
                object.driver.tracks = tracks
                object.driver.load_genre(object, rel=True)

    def clean_node(self, node):
        """
            Cleans a node of all it's children
        """
        iter = self.model.iter_children(node)
        while True:
            if not iter: break
            self.model.remove(iter)
            iter = self.model.iter_children(node)

    def refresh_feed(self, widget, event):
        """
            Refreshes a feed
        """
        selection = self.tree.get_selection()
        (model, iter) = selection.get_selected()
        object = self.model.get_value(iter, 1)
        if isinstance(object, PodcastWrapper):
            self.refresh_podcast(object.path, iter)

    def on_add_podcast(self, widget, event):
        """
            Called when a request to add a podcast is made
        """
        dialog = common.TextEntryDialog(self.exaile.window, _("Enter the location of"
            " the podcast"), _("Add a podcast"))

        if dialog.run() == gtk.RESPONSE_OK:
            name = dialog.get_value()
            dialog.destroy()
            path_id = library.get_column_id(self.db, 'paths', 'name', name)
            if self.db.record_count("podcasts", "path=?", (name, )):
                common.error(self.exaile.window, 
                    _("A podcast with that url already"
                    " exists"))
                return

            self.db.execute("INSERT INTO podcasts( path ) VALUES( ? )",
                (path_id,))
            self.db.commit()

            item = self.model.append(self.podcast,
                [self.track, PodcastWrapper(_("Fetching..."), name)])
            self.tree.expand_row(self.model.get_path(self.podcast), False)

            self.refresh_podcast(name, item)
        dialog.destroy()

    @common.threaded
    def refresh_podcast(self, path, item):
        """
            Refreshes a podcast
        """
        try:
            feed = feedparser.parse(path)
        except IOError:
            gobject.idle_add(common.error, self.exaile.window, 
                _("Could not read feed."))
            return

        gobject.idle_add(self.parse_podcast_xml, path, item, feed)

    def parse_podcast_xml(self, path, iter, feed):
        """
            Parses the xml from the podcast and stores the information to the
            database
        """
        title = feed.feed.title
        description = feed.feed.description
        if not description: description = ""

        try: pub_date = feed.feed.updated
        except AttributeError: pub_date = ""

        # TODO: use the image from the feed as the cover image
        image = ""
        path_id = library.get_column_id(self.db, 'paths', 'name', path)

        self.db.execute("UPDATE podcasts SET title=?, "
            "pub_date=?, description=?, image=? WHERE"
            " path=?", (title, pub_date, description, image, path_id))

        self.model.set_value(iter, 1, PodcastWrapper(title, path))
        root_title = title

        found_items = []
        
        for item in feed['entries']:
            title = item.title 
            link = item.link
            print title, link
            desc = item.description

            desc = self.clean_desc(desc)
            enc = None
            try:
                enc = item.enclosures
            except AttributeError:
                pass

            date = item.updated
            if enc:
                try: size = enc[0].length
                except AttributeError: size = 0
                try: length = enc[0].duration
                except AttributeError: length = 0
                try: loc = str(enc[0].href)
                except AttributeError: loc = ''
            else: continue
            loc_id = library.get_column_id(self.db, 'paths', 'name', loc)
            found_items.append(loc_id)

            row = self.db.read_one("podcast_items", "path", 
                "podcast_path=? AND path=?", (path_id, loc_id))

            # if the feed already exists, don't add it again
            if row: continue

            t = common.strdate_to_time(date)
            date = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(t))

            self.db.update("podcast_items",
                {
                    "podcast_path": path_id,
                    "path": loc_id,
                    "pub_date": date,
                    "description": desc,
                    "size": size,
                    "title": title,
                    "length": length,
                }, "path=? AND podcast_path=?", 
                (loc_id, path_id), row == None)

        # delete all items that aren't still in the feed
        where = ["path!=\"%d\" " % (x) for x in found_items]
        where = " AND ".join(where)
        cur = self.db.cursor()
        cur.execute("DELETE FROM podcast_items WHERE (%s) AND podcast_path=?"
            % (where,), (path_id,))

        self.db.commit()

        gobject.timeout_add(500, self.open_podcast, PodcastWrapper(root_title, path))

    def clean_desc(self, desc):
        """ 
            Cleans description of html, and shortens it to 70 chars
        """
        reg = re.compile("<[^>]*?>", re.IGNORECASE|re.DOTALL)
        desc = reg.sub('', desc)
        desc = re.sub(r"\s+", " ", desc)

        if len(desc) > 70:
            desc = desc[:67] + '...'

        return desc

    def delete_podcast(self, widget, event):
        """ 
            Removes a podcast
        """
        selection = self.tree.get_selection()
        (model, iter) = selection.get_selected()
        object = self.model.get_value(iter, 1)
        path_id = library.get_column_id(self.db, 'paths', 'name', object.path)

        if not isinstance(object, PodcastWrapper): return
        dialog = gtk.MessageDialog(self.exaile.window,
            gtk.DIALOG_MODAL, gtk.MESSAGE_QUESTION, gtk.BUTTONS_YES_NO,
            _("Are you sure you want to delete this podcast?"))
        if dialog.run() == gtk.RESPONSE_YES:
            self.db.execute("DELETE FROM podcasts WHERE path=?", (path_id,))
            self.db.execute("DELETE FROM podcast_items WHERE podcast_path=?", 
                (path_id,))

            self.model.remove(iter)
            self.tree.queue_draw()
            
        dialog.destroy()

    def add_url_to_station(self, item, event):
        """
            Adds a url to an existing station
        """
        selection = self.tree.get_selection()
        (model, iter) = selection.get_selected()
        station = model.get_value(iter, 1)
        radio = library.get_column_id(self.db, 'radio', 'name', str(station))
        
        dialog = common.MultiTextEntryDialog(self.exaile.window,
            _("Add Stream to Station"))
        # TRANSLATORS: Address of a radio station
        dialog.add_field(_("URL:"))
        # TRANSLATORS: Description of a radio station
        dialog.add_field(_("Description:"))
        result = dialog.run()
        if result == gtk.RESPONSE_OK:
            (stream, desc) = dialog.get_values()
            path_id = library.get_column_id(self.db, 'paths', 'name', stream)

            self.db.execute("INSERT INTO radio_items(radio, path, title, "
                "description) VALUES( ?, ?, ?, ?)", 
                (radio, path_id, desc, desc))
            
    def remove_station(self, item, event=None):
        """
            Removes a saved station
        """
        selection = self.tree.get_selection()
        (model, iter) = selection.get_selected()
        name = model.get_value(iter, 1)
        if not isinstance(name, CustomWrapper): return
        name = str(name)

        dialog = gtk.MessageDialog(self.exaile.window, 
            gtk.DIALOG_MODAL, gtk.MESSAGE_QUESTION, gtk.BUTTONS_YES_NO, 
            _("Are you sure you want to permanently delete the selected "
            "station?"))
        result = dialog.run()
        dialog.destroy()
        radio_id = library.get_column_id(self.db, 'radio', 'name', name)

        if result == gtk.RESPONSE_YES:
            
            self.db.execute("DELETE FROM radio WHERE id=?", (radio_id,))
            self.db.execute("DELETE FROM radio_items WHERE radio=?", (radio_id,))
            if library.RADIO.has_key(name):
                del library.RADIO[name]

            self.model.remove(iter)
            self.tree.queue_draw()

    def open_station(self, playlist):
        """
            Opens a station
        """
        all = self.db.select("SELECT title, description, paths.name, bitrate FROM "
            "radio_items,radio,paths WHERE radio_items.radio=radio.id AND "
            "paths.id=radio_items.path AND radio.name=?", (playlist,))

        songs = library.TrackData()
        t = trackslist.TracksListCtrl(self.exaile)
        self.exaile.playlists_nb.append_page(t,
            xlmisc.NotebookTab(self.exaile, playlist, t))
        self.exaile.playlists_nb.set_tab_reorderable(t, True)     
        self.exaile.playlists_nb.set_current_page(
            self.exaile.playlists_nb.get_n_pages() - 1)

        self.exaile.tracks = t

        for row in all:
            info = dict()
            info['artist'] = row[2]
            info['album'] = row[1]
            info['loc'] = row[2]
            info['title'] = row[0]
            info['bitrate'] = row[3]

            track = media.Track()
            track.set_info(**info)
            track.type = 'stream'
            songs.append(track)
        t.playlist = playlist
        t.set_songs(songs)
        t.queue_draw()

    def on_add_station(self, widget):
        """
            Adds a station
        """
        dialog = common.MultiTextEntryDialog(self.exaile.window, 
            _("Add Station"))
        dialog.add_field(_("Station Name:"))
        dialog.add_field(_("Description:"))
        dialog.add_field(_("URL:"))
        response = dialog.run()

        if response == gtk.RESPONSE_OK:
            (name, desc, url) = dialog.get_values()
            if not name or not url:
                common.error(self.exaile.window, _("The 'Name' and 'URL'"
                    " fields are required"))
                self.on_add_station(widget)
                return

            c = self.db.record_count("radio", "name=?", (name,))

            if c > 0:
                common.error(self.exaile.window, _("Station name already exists."))
                return
            radio_id = library.get_column_id(self.db, 'radio', 'name', name)
            path_id = library.get_column_id(self.db, 'paths', 'name', url)
            self.db.execute("INSERT INTO radio_items(radio, path, title, "
                "description) VALUES( ?, ?, ?, ? )", (radio_id, path_id, desc,
                desc))
            
            item = self.model.append(self.custom, [self.track, 
                CustomWrapper(name)])
            path = self.model.get_path(self.custom)
            self.tree.expand_row(path, False)

    def on_add_to_station(self, widget, event):
        """
            Adds a playlist to the database
        """
        dialog = common.TextEntryDialog(self.exaile.window,
            _("Enter the name of the station"),
            _("Enter the name of the station"))
        result = dialog.run()
        if result == gtk.RESPONSE_OK:
            name = dialog.get_value()
            if name == "": return
            c = self.db.record_count("radio", "name=?",
                (name,))

            if c > 0:
                common.error(self, _("Station already exists."))
                return

            station_id = library.get_column_id(self.db, 'radio', 
                'name', name)
            
            self.model.append(self.custom, [self.track, CustomWrapper(name)])
            self.tree.expand_row(self.model.get_path(self.custom), False)

            self.add_items_to_station(station=name)

    def add_items_to_station(self, item=None, event=None, 
        ts=None, station=None):
        """
            Adds the selected tracks tot he playlist
        """

        if ts == None: ts = self.exaile.tracks
        songs = ts.get_selected_tracks()

        if station:
            playlist = station
        else:
            playlist = unicode(item.get_child().get_text(), 'utf-8')

        station_id = library.get_column_id(self.db, 'radio', 'name', playlist)

        for track in songs:
            if track.type != 'stream': continue
            path_id = library.get_column_id(self.db, 'paths', 'name', track.loc)
            try:
                album = track.album
                if not album: album = track.artist
                self.db.execute("INSERT INTO radio_items( radio, title, path, "
                    "description, bitrate ) " \
                    "VALUES( ?, ?, ?, ?, ? )",
                    (station_id, track.title, path_id,
                    album, track.bitrate))
            except sqlite.IntegrityError:
                pass
        self.db.commit()

